package ru.dbotthepony.mc.otm.item;

import net.minecraft.ChatFormatting;
import net.minecraft.core.Direction;
import net.minecraft.nbt.CompoundTag;
import net.minecraft.network.chat.Component;
import net.minecraft.network.chat.TranslatableComponent;
import net.minecraft.world.item.*;
import net.minecraft.world.level.Level;
import net.minecraftforge.common.capabilities.Capability;
import net.minecraftforge.common.capabilities.ICapabilityProvider;
import net.minecraftforge.common.util.LazyOptional;
import ru.dbotthepony.mc.otm.capability.IMatterHandler;
import ru.dbotthepony.mc.otm.capability.MatteryCapability;
import ru.dbotthepony.mc.otm.menu.FormattingHelper;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.math.BigDecimal;
import java.util.List;
import java.util.Objects;

public class ItemMatterCapacitor extends Item {
	public static class ItemMatterCapacitorCapability implements ICapabilityProvider, IMatterHandler {
		private final LazyOptional<IMatterHandler> resolver = LazyOptional.of(() -> this);

		@Nonnull
		@Override
		public <T> LazyOptional<T> getCapability(@Nonnull Capability<T> cap, @Nullable Direction side) {
			if (cap == MatteryCapability.MATTER)
				return resolver.cast();

			return LazyOptional.empty();
		}
		private final BigDecimal storage;
		private final BigDecimal max_input;
		private final BigDecimal max_output;
		private final boolean is_creative;
		private BigDecimal energy;
		private String _energy;

		private final ItemStack stack;

		private BigDecimal matter() {
			CompoundTag tag = stack.getOrCreateTag();

			if (tag.contains("otm_matter")) {
				String get_energy = tag.getString("otm_matter");

				if (Objects.equals(_energy, get_energy)) {
					return energy;
				}

				energy = new BigDecimal(tag.getString("otm_matter"));
				_energy = get_energy;

				return energy;
			}

			return BigDecimal.ZERO;
		}

		private void matter(BigDecimal value) {
			energy = value;
			stack.getOrCreateTag().putString("otm_matter", value.toString());
		}

		public ItemMatterCapacitorCapability(ItemStack stack, BigDecimal storage, BigDecimal max_input, BigDecimal max_output) {
			this.stack = stack;
			is_creative = false;
			this.storage = storage;
			this.max_input = max_input;
			this.max_output = max_output;
		}

		public ItemMatterCapacitorCapability(ItemStack stack) {
			this.stack = stack;
			is_creative = true;
			storage = MatteryCapability.LONG_MAX_VALUE;
			max_input = MatteryCapability.LONG_MAX_VALUE;
			max_output = MatteryCapability.LONG_MAX_VALUE;
		}

		public ItemMatterCapacitorCapability(ItemStack stack, BigDecimal storage) {
			this.stack = stack;
			is_creative = false;
			this.storage = storage;
			max_input = MatteryCapability.LONG_MAX_VALUE;
			max_output = MatteryCapability.LONG_MAX_VALUE;
		}

		@Nonnull
		@Override
		public BigDecimal getStoredMatter() {
			if (is_creative)
				return MatteryCapability.LONG_MAX_VALUE;

			return matter();
		}

		@Nonnull
		@Override
		public BigDecimal getMaxStoredMatter() {
			if (is_creative)
				return MatteryCapability.LONG_MAX_VALUE;

			return storage;
		}

		@Nonnull
		@Override
		public BigDecimal getMissingMatter() {
			if (is_creative)
				return MatteryCapability.LONG_MAX_VALUE;

			return IMatterHandler.super.getMissingMatter();
		}

		@Nonnull
		@Override
		public BigDecimal receiveMatterOuter(BigDecimal howMuch, boolean simulate) {
			return receiveMatterInner(howMuch, simulate);
		}

		@Nonnull
		@Override
		public BigDecimal receiveMatterInner(BigDecimal howMuch, boolean simulate) {
			if (is_creative)
				return howMuch;

			BigDecimal new_matter = matter().add(howMuch.min(max_input), MatteryCapability.ROUND_RULES).min(storage);
			BigDecimal diff = new_matter.subtract(matter());

			if (!simulate) {
				matter(new_matter);
			}

			return diff;
		}

		@Nonnull
		@Override
		public BigDecimal extractMatterOuter(BigDecimal howMuch, boolean simulate) {
			return extractMatterInner(howMuch, simulate);
		}

		@Nonnull
		@Override
		public BigDecimal extractMatterInner(BigDecimal howMuch, boolean simulate) {
			if (is_creative)
				return howMuch;


			BigDecimal new_matter = matter().subtract(howMuch.min(max_output), MatteryCapability.ROUND_RULES).max(BigDecimal.ZERO);
			BigDecimal diff = matter().subtract(new_matter);

			if (!simulate) {
				matter(new_matter);
			}

			return diff;
		}

		@Nonnull
		@Override
		public MatterDirection getDirection() {
			return MatterDirection.BIDIRECTIONAL;
		}
	}

	public final BigDecimal storage;

	private static final Component INFINITE_STORAGE = new TranslatableComponent("otm.item.matter.infinite").withStyle(ChatFormatting.GRAY);

	private final boolean is_creative;

	public ItemMatterCapacitor(BigDecimal storage) {
		super(new Properties().stacksTo(1).tab(CreativeModeTab.TAB_MISC));
		is_creative = false;
		this.storage = storage;
	}

	public ItemMatterCapacitor() {
		super(new Properties().stacksTo(1).tab(CreativeModeTab.TAB_MISC).rarity(Rarity.EPIC));
		is_creative = true;
		storage = MatteryCapability.LONG_MAX_VALUE;
	}

	@Override
	public void appendHoverText(ItemStack stack, @Nullable Level p_41422_, List<Component> p_41423_, TooltipFlag p_41424_) {
		super.appendHoverText(stack, p_41422_, p_41423_, p_41424_);

		if (is_creative) {
			p_41423_.add(INFINITE_STORAGE);
		} else {
			if (stack.getCapability(MatteryCapability.MATTER).resolve().get() instanceof ItemMatterCapacitorCapability cap)
				p_41423_.add(new TranslatableComponent("otm.item.matter.normal", FormattingHelper.formatMatterValuePlain(cap.getStoredMatter()), FormattingHelper.formatMatterValuePlain(storage)).withStyle(ChatFormatting.GRAY));
			else
				p_41423_.add(new TranslatableComponent("otm.item.matter.normal", "???", FormattingHelper.formatMatterValuePlain(storage)).withStyle(ChatFormatting.GRAY));
		}
	}

	@Nullable
	@Override
	public ICapabilityProvider initCapabilities(ItemStack stack, @Nullable CompoundTag nbt) {
		if (is_creative)
			return new ItemMatterCapacitorCapability(stack);

		return new ItemMatterCapacitorCapability(stack, storage);
	}
}
